/* ----------------------------------------------------------------------------
 *     Copyright (c) 2000-2021, Nexus Electronics Ltd (www.nexus.co.uk)
 *                            All rights reserved
 *
 * Redistribution and use in source and derived forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in derived forms must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  * Neither the name of Nexus Electronics Ltd. nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY NEXUS ELECTRONICS LTD. "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NEXUS ELECTRONICS LTD. BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * ----------------------------------------------------------------------------
 * Title       : triple-DES, top level
 * Project     : MAS
 * RCS ID      : 
 * $Id: triple_des.v,v 1.2 2001/01/16 18:13:52 el Exp $
 * ----------------------------------------------------------------------------
 * File        : triple_des.v
 * Author      : E.M. Lavelle
 * Company     : Nexus Electronics Ltd.
 * Last saved  : 22/08/2000
 * Version     : $Revision: 1111 $
 * ----------------------------------------------------------------------------
 * Description : 
 * 
 * top-level module, triple-DES. the module has a 64-bit input bus (which 
 * loads one of 5 nominal input registers), and a 64-bit output bus. the
 * input bus is used to load the IV, keys, and the data (the plaintext for
 * encoding or the ciphertext for decoding). the result (encoded ciphertext
 * or decoded plaintext) is presented on the 64-bit ouput bus. the output
 * bus comes direct from a register; the input bus is loaded directly to a
 * register, and to a register file. the remaining inputs may be combined
 * before being registered.
 * 
 * all inputs are sampled when DAV_H is active. the required load register
 * is given by the ADDR inputs. the outputs are valid only when DRDY_H is
 * sampled active.
 * 
 * the required operation is given by the ENC3D_H and CBC_H control inputs:
 * ENC3D_H
 *   CBC_H
 * 0 0      ECB mode, decode
 * 0 1      CBC mode, decode
 * 1 0      ECB mode, encode
 * 1 1      CBC mode, encode
 * note that the control inputs are only sampled when DAV_H is active, and
 * the current cycle is a data register load.
 * 
 * the input register to load is given by the ADDR inputs:
 * ADDR: 000      Key 0
 *       001      Key 1
 *       010      Key 2
 *       011      IV
 *       100      data register (plaintext/ciphertext)
 *       others   reserved; writes ignored
 * 
 * the unit starts an encrypt or decrypt operation when the data register
 * is loaded. the output must be latched when DRDY_H is found active;
 * currently, this occurs 91 clock cycles later. operations cannot be
 * pipelined. 2-key operation can be achieved by making key1 equal to key3.
 * if there are any load operations between the data register load, and the
 * time at which DRDY_H is returned, then there is a possibility that the
 * output data will be incorrect.
 * 
 * when the IV register is loaded, it applies only to the next encode or
 * decode operation (for CBC mode; it is ignored for ECB mode). the standard
 * chaining mechanism is used for subsequent encodes or decodes, until the
 * next IV load. in other words, the IV load specifies the start of a block
 * in CBC mode. note that the IV data is lost after it is used, and there is
 * currently no mechanism to automatically update the IV for subsequent blocks
 * (such as an IV incrementer). this is largely because I don't understand how
 * the IV will be used in practice - the IV in a real application could be
 * actually be a random data word, for example.
 * 
 * the registers must be loaded in the following order, at a time when the
 * unit is not running:
 * 1) the keys, if they have changed;
 * 2) the IV, if it is required (even if it is unchanged);
 * 3) the data register, which starts the required op.
 * loads may occur on adjacent clock cycles, with no delays.
 * 
 * the ubit has a sync reset, but not an async reset. note that the ordering
 * of most internal buses is given as [1:n], which is consistent with
 * standard DES documentation.
 * 
 * ----------------------------------------------------------------------------
 * Modification history:
 * Date        Version Author     Comment
 * 22/08/2000  0.1      EML       created
 *
 * ------------------------------------------------------------------------- */

`include "3des_inc.v"           // register decodes

module triple_des(
	   CLK,                 // clock
	   SRST_H,              // sync reset
	   ICBC_H,              // ECB/CBC mode
	   IENC3D_H,            // encrypt/decrypt
	   DAV_H,               // data available strobe
	   ADDR,                // input register address
	   DATAIN,              // input data
	   DRDY_H,              // data ready strobe
	   DATAOUT              // 64-bit output
	   );
   input         CLK, SRST_H, DAV_H, ICBC_H, IENC3D_H;
   input [2:0] 	 ADDR;
   input [1:64]  DATAIN;
   output 	 DRDY_H; 	 
   output [1:64] DATAOUT;
   
   // internal variables
   reg [1:64] 	 R1, R2, R3;
   reg [1:32] 	 R4_LEFT, R4_RIGHT, R4_FB_RIGHT;
   reg 		 CBC_H, ENC3D_H, DAV2_H;
   wire [1:64] 	 PLAIN_IP;
   wire [1:32] 	 R1_LEFT, R1_RIGHT, R3_LEFT, R3_RIGHT;
   wire 	 START_H, R3LE_H;

   // module connections
   wire [1:48] 	 PRTL_KEY;
   wire [1:32] 	 PBOX;
   wire [1:0] 	 KSEL;
   wire 	 CBCENC_H, CBCDEC_H, ENC_H, DRDY1_H, DRDY2_H, DLD_H, SWAP_H, 
		 DCE_H, KLD_H;

   // name changes 
   assign DRDY_H   = DRDY2_H;
   assign R3_LEFT  = R3[ 1:32];
   assign R3_RIGHT = R3[33:64];
   assign R1_LEFT  = R1[ 1:32];
   assign R1_RIGHT = R1[33:64];

   // start condition (data register load)
   assign START_H = DAV_H & (ADDR == `REG_RUN);
   
   // register the control inputs when starting
   always @(posedge CLK) begin
      if(SRST_H) begin		// sync reset
	 CBC_H    <= 0;
	 ENC3D_H  <= 0;
      end
      else if(START_H) begin	// load enable
	 CBC_H    <= ICBC_H;
	 ENC3D_H  <= IENC3D_H;
      end
   end

   // DAV_H must be delayed one cycle for the CBC encode IV load
   always @(posedge CLK) begin
      if(SRST_H)
	DAV2_H <= 0;
      else
	DAV2_H <= DAV_H;
   end
   
   // CBC encode and decode conditions. these are don't touched to ensure
   // that CLBs down the line have a reasonable number of inputs

   // synthesis attribute keep of CBCENC_H is "true"
   // synthesis attribute keep of CBCDEC_H is "true"
   assign CBCENC_H = CBC_H &  ENC3D_H;
   assign CBCDEC_H = CBC_H & ~ENC3D_H;

   // instantiate the controller
   controller U1(CLK, SRST_H, START_H, ENC3D_H, KSEL, KLD_H, DLD_H,
		 DCE_H, SWAP_H, ENC_H, DRDY1_H, DRDY2_H);

   // instantiate the key generator
   key U2(CLK, SRST_H, DAV_H, ADDR, DATAIN, ENC_H, KSEL, KLD_H, PRTL_KEY);

   // instantiate the round processing
   desround U3(R4_RIGHT, PRTL_KEY, PBOX);

   // IP: initial permutation
   assign PLAIN_IP = 
	  {DATAIN[58], DATAIN[50],  DATAIN[42], DATAIN[34],
	   DATAIN[26], DATAIN[18],  DATAIN[10], DATAIN[2],
	   DATAIN[60], DATAIN[52],  DATAIN[44], DATAIN[36],
	   DATAIN[28], DATAIN[20],  DATAIN[12], DATAIN[4],
	   DATAIN[62], DATAIN[54],  DATAIN[46], DATAIN[38],
	   DATAIN[30], DATAIN[22],  DATAIN[14], DATAIN[6],
	   DATAIN[64], DATAIN[56],  DATAIN[48], DATAIN[40],
	   DATAIN[32], DATAIN[24],  DATAIN[16], DATAIN[8],
	   DATAIN[57], DATAIN[49],  DATAIN[41], DATAIN[33],
	   DATAIN[25], DATAIN[17],  DATAIN[9],  DATAIN[1],
	   DATAIN[59], DATAIN[51],  DATAIN[43], DATAIN[35],
	   DATAIN[27], DATAIN[19],  DATAIN[11], DATAIN[3],
	   DATAIN[61], DATAIN[53],  DATAIN[45], DATAIN[37],
	   DATAIN[29], DATAIN[21],  DATAIN[13], DATAIN[5],
	   DATAIN[63], DATAIN[55],  DATAIN[47], DATAIN[39],
	   DATAIN[31], DATAIN[23],  DATAIN[15], DATAIN[7]};

   // R1: input register. this has to load for all DAV_H's. note that the
   // IP occurs *before* the register (the IP doesn't correspond to any
   // hardware: it's just a bus swap)
   always @(posedge CLK) begin
      if(DAV_H)                      // load enable
	R1 <= PLAIN_IP;
   end

   // R2: store the previous ciphertext input for CBC decode. note that the
   // IV load is automatic, as long as there are no key loads between the
   // IV and the data.
   always @(posedge CLK) begin
      if(DAV_H)                      // load enable 
	R2 <= R1;
   end
   
   /* R3: data/ciphertext output register. this has 3 functions:
    * 1  holds the encode IV. to do this, it loads R2 one cycle after
    *    a DAV_H; otherwise:
    * 2  for CBC decryption: load R2 ^ R4 (ie. xor previous plaintext
    *    with current ciphertext), otherwise
    * 3  all other modes (CBC encrypt, and both ECB): load R4
    * if the don't touch on the load enable works, then the logic should
    * fit in 32 slices. there will be an extra 32 slices if not.
    */
   assign R3LE_H = DAV2_H | DRDY1_H /* synthesis dont touch */;
   always @(posedge CLK) begin
      if(R3LE_H) begin		// load enable, hopefully
	 if(DAV2_H)		// encode IV load
	   R3 <= R2;
	 else if(CBCDEC_H)	// CBC decode: XOR operation
	   R3 <= R2 ^ {R4_LEFT, R4_RIGHT};
	 else			// other modes: pass R4
	   R3 <= {R4_LEFT, R4_RIGHT};
      end
   end
   
   // output permutation
   assign DATAOUT =
	  {R3[40], R3[8],  R3[48], R3[16], R3[56], R3[24], R3[64], R3[32],
	   R3[39], R3[7],  R3[47], R3[15], R3[55], R3[23], R3[63], R3[31],
	   R3[38], R3[6],  R3[46], R3[14], R3[54], R3[22], R3[62], R3[30],
	   R3[37], R3[5],  R3[45], R3[13], R3[53], R3[21], R3[61], R3[29],
	   R3[36], R3[4],  R3[44], R3[12], R3[52], R3[20], R3[60], R3[28],
	   R3[35], R3[3],  R3[43], R3[11], R3[51], R3[19], R3[59], R3[27],
	   R3[34], R3[2],  R3[42], R3[10], R3[50], R3[18], R3[58], R3[26],
	   R3[33], R3[1],  R3[41], R3[9],  R3[49], R3[17], R3[57], R3[25]};
   
   // -----------------------------------------------------------------------
   // R4 load. the load data to the R register is a function of 8
   // inputs, and 6 inputs for the L register. to handle this, I 
   // generate an intermediate signal, CBCENC_H, combining CBC_H 
   // and ENC3D_H. the right input can then be implemented in 3 LUTs,
   // and the left input in 2 LUTs, giving 160 LUTs.
   // -----------------------------------------------------------------------

   // first LUT on right input: in pass mode, either the PBOX output is
   // loaded back into the right-half register, or the left half is loaded
   // in if a swap is required
   always @(SWAP_H or R4_LEFT or PBOX) begin
      if(SWAP_H)
	R4_FB_RIGHT = R4_LEFT;
      else
	R4_FB_RIGHT = R4_LEFT ^ PBOX;
   end

   // 2nd part of R4_RIGHT loading: 5 inputs, 2 LUTs
   always @(posedge CLK) begin
      if(DCE_H) begin                 // clock enable 
	 case ({DLD_H, CBCENC_H})
	   2'b00   : R4_RIGHT <= R4_FB_RIGHT;
	   2'b01   : R4_RIGHT <= R4_FB_RIGHT;
	   2'b10   : R4_RIGHT <= R1_RIGHT;
	   default : R4_RIGHT <= R3_RIGHT ^ R1_RIGHT;
	 endcase
      end
   end

   // R4_LEFT: 5 inputs, 2 LUTs
   always @(posedge CLK) begin
      if(DCE_H) begin                 // clock enable 
	 case ({DLD_H, CBCENC_H})
	   2'b00   : R4_LEFT <= R4_RIGHT;
	   2'b01   : R4_LEFT <= R4_RIGHT;
	   2'b10   : R4_LEFT <= R1_LEFT;
	   default : R4_LEFT <= R3_LEFT ^ R1_LEFT;
	 endcase
      end 
   end
   
endmodule // triple_des

/*----------------------------------- EOF -----------------------------------*/
